@@ -31,16 +31,18 @@ import ctypes
 import os
 import sys
 
-from typing import Sequence, Final, Union, Optional, Dict, cast
+from typing import Sequence, Final, Union, Optional, Dict, Any, cast
 
 from . import mixin
 
 from .util import MaybeSequence
-from .meta import halcon_native_lib_names
+from .meta import halcon_native_lib_names, expected_halcon_version
 
+import halcon
 
 __all__ = [
     'get_sem_type',
+    'H_MSG_FAIL',
     'H_MSG_OK',
     'HalconOperator',
     'HCkP',
@@ -77,6 +79,24 @@ class HOperatorError(HError):
     def __init__(self, error_code: int):
         self.error_code = error_code
         self.message = _get_error_text(error_code)
+        self.extended_error_message = ''
+
+    def __str__(self) -> str:
+        if self.extended_error_message == '':
+            return self.message
+
+        return f'{self.message}: {self.extended_error_message}'
+
+    def _update_extended_info(self) -> None:
+        try:
+            _, extended_code, extended_msg = halcon.get_extended_error_info()
+            self.extended_error_code = extended_code
+            self.extended_error_message = extended_msg
+        except HOperatorError as exc:
+            sys.stderr.write(
+                '[WARNING] failed to call get_extended_error_info: '
+                f'{exc}'
+            )
 
 
 class HTupleConversionError(HError):
@@ -103,10 +123,18 @@ class HHandleBase(object):
         HCkP(_halcon_lib.HLIClearHandle(self._handle_ptr))
 
 
-# TODO what should what shouldn't be valid with HNull?
+# TODO what should and what shouldn't be valid with HNull?
 class HNull(HHandleBase):
     """HALCON HNULL handle class."""
-    pass
+    def __init__(self):
+        self._handle_ptr = ctypes.c_void_p(0)
+
+    def __del__(self) -> None:
+        pass
+
+    def __eq__(self, other: Any) -> bool:
+        # HNull has no state.
+        return isinstance(other, HNull)
 
 
 # --- Type Definitions ---
@@ -124,6 +152,7 @@ Hkey = ctypes.c_void_p
 
 # TODO automatic constant and error code generation.
 H_MSG_OK = 2
+H_MSG_FAIL = 5
 
 
 class HObjectBase(object):
@@ -245,7 +274,8 @@ class HalconOperator(object):
 
     def init_oct(self, par_index: int) -> None:
         """Initialize output control tuple."""
-        # TODO extend HLI to allow init multiple at once, to avoid FFI roundtrips.
+        # TODO extend HLI to allow init multiple at once,
+        # to avoid FFI roundtrips.
         HCkP(_halcon_lib.HLIInitOCT(
             self.proc_handle,
             ctypes.c_int32(par_index)
@@ -328,11 +358,15 @@ class HalconOperator(object):
         # TODO maybe only if already called and has proc_result.
         HCkP(_halcon_lib.HLIDestroyProcedure(
             self.proc_handle,
-            # Either execute result or error from setup calls, or H_MSG_OK,
-            # if something else went wrong, like tuple convert error.
+            # TODO more research is needed here, it's still unclear if this
+            # isn't wrong.
             ctypes.c_int32(proc_result)
         ))
 
+        if exc_type == HOperatorError:
+            # Has to happen after HLIDestroyProcedure.
+            exc_value._update_extended_info()
+
 
 # --- Exported Functions ---
 
@@ -367,6 +401,24 @@ def load_native_lib(names: Dict[str, str]) -> ctypes.CDLL:
     else:
         raise HError(f'Unsupported operating system {sys.platform}')
 
+    lib = _dlopen(name)
+
+    return lib
+
+
+def make_c_func_prototype(restype, *argtypes):
+    if _is_32bit_win:
+        return ctypes.WINFUNCTYPE(restype, *argtypes)
+
+    return ctypes.CFUNCTYPE(restype, *argtypes)
+
+
+# --- Private Implementation Details ---
+
+_is_32bit_win = os.name == 'nt' and sys.maxsize == (2**31 - 1)
+
+
+def _dlopen(name) -> ctypes.CDLL:
     # Python 3.8 added the winmode parameter. If it is not specified, Python
     # will attempt to open the DLL using LoadLibraryEx's
     # LOAD_LIBRARY_SEARCH_DEFAULT_DIRS mode, which does not take the PATH
@@ -383,19 +435,187 @@ def load_native_lib(names: Dict[str, str]) -> ctypes.CDLL:
     return lib
 
 
-def make_c_func_prototype(restype, *argtypes):
-    if _is_32bit_win:
-        return ctypes.WINFUNCTYPE(restype, *argtypes)
-
-    return ctypes.CFUNCTYPE(restype, *argtypes)
+def _is_compatible_version(got: str) -> bool:
+    return got.rpartition('.')[0] == expected_halcon_version.rpartition('.')[0]
 
 
-# --- Private Implementation Details ---
+def _is_compatible_mac_version(got: str) -> bool:
+    # Mac HALCON_FRAMEWORK_VERSION eg 21.11P(progress) 21.11(steady)
+    return got.replace('P', '') == expected_halcon_version.replace('P', '')
+
+
+def _halcon_version_to_semver(version: str) -> str:
+    return version.replace('.', '')
+
+
+def _log_compatibility_warning(version: str):
+    package_version = _halcon_version_to_semver(version)
+    sys.stderr.write(
+        '[WARNING] Wrong interface package version.\n'
+        f'Expected {expected_halcon_version} but found {version}\n'
+        'Compatibility is not guaranteed and crashes etc. are possible.\n'
+        'Please install the matching interface package version: '
+        f'"mvtec-halcon=={package_version}"\n'
+    )
+
+
+def _incompatible_versions(potential_version: str) -> str:
+    package_version = _halcon_version_to_semver(potential_version)
+    return (
+        'Incompatible HALCON library versions. '
+        f'Expected version {expected_halcon_version} but found: '
+        f'{potential_version}.\n'
+        f'Either change LD_LIBRARY_PATH to point to a HALCON '
+        f'{expected_halcon_version} installation, '
+        f'or install the matching interface package version: '
+        f'"mvtec-halcon=={package_version}"'
+    )
+
+
+def _compat_load_halcon_linux() -> ctypes.CDLL:
+    name = halcon_native_lib_names['linux']
+    try:
+        return _dlopen(name['base'].format(name['version']))
+    except OSError:
+        try:
+            # If there is any HALCON version dynamically loadable this links
+            # to it.
+            # Compatibility will be checked by _halcon_version_check.
+            return _dlopen('libhalcon.so')
+        except OSError as exc:
+            if 'No such file' in str(exc):
+                raise HError(
+                    'Unable to find any HALCON library.\n'
+                    'Set LD_LIBRARY_PATH in your environment, for example:\n'
+                    'export LD_LIBRARY_PATH=$HALCONROOT/lib/x64-linux'
+                ) from None
+            else:
+                # There are other error that we don't want to mask.
+                raise exc from None
+
+
+def _compat_load_halcon_darwin() -> ctypes.CDLL:
+    name = halcon_native_lib_names['darwin']
+    try:
+        return _dlopen(name['base'].format(name['version']))
+    except OSError:
+        try:
+            # If there is any HALCON version dynamically loadable this links
+            # to it.
+            # Compatibility will be checked by _halcon_version_check.
+            return _dlopen(name['base'].format('Current'))
+        except OSError as exc:
+            if 'image not found' in str(exc):
+                raise HError(
+                    'Unable to find any HALCON library.\n'
+                    'Set DYLD_FRAMEWORK_PATH in your environment.'
+                ) from None
+            else:
+                raise exc from None
+
+
+def _compat_load_win() -> ctypes.CDLL:
+    try:
+        # This name is already non version specific, so there is no fallback
+        # we could try.
+        return _dlopen(halcon_native_lib_names['win'])
+    except FileNotFoundError:
+        # Only on windows we get a more specific error FileNotFoundError
+        # instead of OSError. On the other platforms the error string is the
+        # only source we have to differentiate the errors.
+        raise HError(
+            'Unable to find any HALCON library.\n'
+            'Set PATH in your environment, for example:\n'
+            'set PATH="$HALCONROOT/bin/x64-win64;%PATH%"'
+        ) from None
+
+
+def _compat_load_halcon() -> ctypes.CDLL:
+    """
+    Dynamically load native HALCON library with compatibility shim.
+
+    Returns
+    -------
+
+    lib : ctypes.CDLL
+          Native library handle.
+
+    Notes
+    -----
+    Due to our requirement of having to find the correct dynamic library on
+    Linux and Mac we have to have the full version string information on the
+    Python side, instead of realxing the version requirements on the package
+    side, we chose to give better and more helpful error messages and use a
+    heuristic to find and use possibly incompatible HALCON dynamic libraries
+    as follows:
+
+    Users can end up with a version mismatch by doing:
+
+    pip install (no version):
+        -> HALCON major too new
+        -> wrong steady / progress
+
+    pip install (wrong version):
+        -> HALCON major too old
+        -> HALCON major too new
+        -> wrong steady / progress
+
+    We have no control about pip version resolution logic.
+    We can catch the error when the package is imported, and library is
+    loaded as a result.
+
+    HALCON major too old:
+    Hard error
+
+    HALCON major too new:
+    Hard error
+
+    wrong steady / progress:
+    Try search for compatible 2111X X: 0-9
+    WARNING in stderr: Might be broken but could work:
+    Expected package version X but got Y please install Y.
+
+    The logic on each platfrom is a bit different.
+
+    Technically this has TOCTTOU problems in the fallback mode,
+    but they seem acceptable for the situation.
+    """
+    if sys.platform.startswith('linux'):
+        return _compat_load_halcon_linux()
+    elif sys.platform.startswith('darwin'):
+        return _compat_load_halcon_darwin()
+    elif sys.platform == 'win32' or sys.platform == 'cygwin':
+        return _compat_load_win()
+    else:
+        raise HError(f'Unsupported operating system {sys.platform}')
+
+
+def _halcon_version_check(halcon_lib: ctypes.CDLL):
+    # For example on Windows its possible to have loaded an incompatible
+    # HALCON library.
+    if hasattr(halcon_lib, 'HLIVersion'):
+        out_c_str = ctypes.c_char_p()
+        halcon_lib.HLIVersion(ctypes.byref(out_c_str))
+        halcon_version = out_c_str.value.decode('utf-8')
+        if halcon_version != expected_halcon_version:
+            if _is_compatible_version(halcon_version):
+                _log_compatibility_warning()
+            else:
+                raise HError(_incompatible_versions(halcon_version))
+    else:
+        # Hlib versions that are too old.
+        # HLIVersion was added with 21.05.
+        raise HError(
+            'Incompatible HALCON library version.\n'
+            f'Expected version: {expected_halcon_version}.\n'
+            'Configure your environment so that the correct version '
+            'can be found.'
+        )
 
-_is_32bit_win = os.name == 'nt' and sys.maxsize == (2**31 - 1)
 
 # Load dynamic HALCON library.
-_halcon_lib: Final[ctypes.CDLL] = load_native_lib(halcon_native_lib_names)
+_halcon_lib: Final[ctypes.CDLL] = _compat_load_halcon()
+_halcon_version_check(_halcon_lib)
 
 
 class _HTuple(object):
@@ -410,14 +630,23 @@ class _HTuple(object):
     This is ONLY meant for internal use.
     """
 
-    def __init__(self, value: Optional[HTupleType]):
+    # Funky looking signature to avoid accidental user None passing through.
+    def __init__(
+        self,
+        value: Optional[HTupleType],
+        empty: Optional[int] = None
+    ):
         """Construct HTuple with value."""
         self._tuple_ptr = ctypes.c_ssize_t()
         HCkP(_halcon_lib.HLICreateTuple(ctypes.byref(self._tuple_ptr)))
 
-        if value is not None:
+        if empty != 1:
             _python_to_htuple(self._tuple_ptr, value)
 
+    @staticmethod
+    def new_empty() -> '_HTuple':
+        return _HTuple(value=None, empty=1)
+
     def as_python(self, as_list: bool) -> HTupleType:
         """Convert native HTuple to python object."""
         return _htuple_to_python(self._tuple_ptr, as_list)
@@ -437,7 +666,6 @@ def _get_error_text(error_code: int) -> str:
     # Keep in sync with #define MAX_STRING 1024
     out_string = ctypes.create_string_buffer(1024)
     # Returns void should not fail.
-    # TODO extended error messages.
     _halcon_lib.HLIGetErrorMessage(error_code, out_string)
     return out_string.value.decode('utf-8')
 
@@ -470,10 +698,7 @@ def _set_htuple_value(
         assert _is_valid_ssize_t(int_value)
         HCkP(_halcon_lib.HLISetL(tuple_ptr, index, ctypes.c_int64(int_value)))
     elif isinstance(value, HHandleBase):
-        if isinstance(value, HNull):
-            HCkP(_halcon_lib.HLISetH(tuple_ptr, index, 0))
-        else:
-            HCkP(_halcon_lib.HLISetH(tuple_ptr, index, value._handle_ptr))
+        HCkP(_halcon_lib.HLISetH(tuple_ptr, index, value._handle_ptr))
     elif value_type == str:
         c_string = ctypes.c_char_p(cast(str, value).encode('utf-8'))
         HCkP(_halcon_lib.HLISetS(tuple_ptr, index, c_string))
@@ -575,7 +800,7 @@ def _htuple_element_to_python(
         ))
 
         if out_handle_ptr.value is None:
-            return HNull(ctypes.c_void_p())
+            return HNull()
 
         return mixin.to_handle(out_handle_ptr)
 
